#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to simulate a watercolor sketched painting

# (c) Issabella (tutorial) / Dinasset (filter) / 2021-2024
#
#   History:
#   ------------------------------------------------------------------
#   2021-08-30: Initial version as JVID_L/LA
#   2024-02-27: Rewamped version, largerly modified...
#               ...hence decided to apply a new name
#   ------------------------------------------------------------------
#   v.1.1.0: 2024-February initial version as PlainKnittedLook2024
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 
from gimpfu import *
import decimal
import random
import time
import datetime

# Service routines   
def get_saturation (layer):

    image = pdb.gimp_item_get_image (layer)
    sat_image = pdb.plug_in_decompose_registered (image, layer, "Saturation", False )
    #decompose returns an array, in this case of 1 element
    sat_layer = pdb.gimp_image_get_active_layer (sat_image[0]) #element 0 is the only one returned
    saturation = pdb.gimp_histogram (sat_layer, HISTOGRAM_VALUE, 0, 255 )
    pdb.gimp_image_delete (sat_image[0])
    return(saturation[0])

def get_value(layer):

    image = pdb.gimp_item_get_image (layer)
    val_image = pdb.plug_in_decompose_registered (image, layer, "Value", False )
    #decompose returns an array, in this case of 1 element
    val_layer = pdb.gimp_image_get_active_layer (val_image[0]) #element 0 is the only one returned
    value = pdb.gimp_histogram (val_layer, HISTOGRAM_VALUE, 0, 255 )
    pdb.gimp_image_delete (val_image[0])
    return(value[0])
     
# ------------------- 
# convert RGB to HSV 
# ------------------- 

def rgb2hsv(rgb):
    (r,g,b,a) = rgb
    r, g, b = r/255.0, g/255.0, b/255.0
    mx = max(r, g, b)
    mn = min(r, g, b)
    df = mx-mn
    if mx == mn:
        h = 0
    elif mx == r:
        h = (60 * ((g-b)/df) + 360) % 360
    elif mx == g:
        h = (60 * ((b-r)/df) + 120) % 360
    elif mx == b:
        h = (60 * ((r-g)/df) + 240) % 360
    if mx == 0:
        s = 0
    else:
        s = (df/mx)*100
    v = mx*100
    #info_msg ("hue="+str(h))
    #info_msg ("sat="+str(s))
    #info_msg ("val="+str(v))
    return h, s, v

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
############################################################################
#
def PlainKnittedLook2024 (inImage, inDrawable, widthheight,
                            nrAreas, patname,   #2024 added pattern-name
                            frameamt, framecol, #2024 added 
                            doFlatten,
                            ) :

    pdb.gimp_context_push()
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    width = inImage.width
    height = inImage.height
    itype = inImage.base_type
    
    name = inDrawable.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H-%M-%S')
    
    # calculate scale factor for the new output image
    if widthheight == 0 :   # means the user wants to have same size as input
        widthheight = width+height
    # check/adjust size not smaller than 2000 as w+h
    if widthheight < 2000 :   # not too small, pls
        widthheight = 2000
    
    optiWidthHeight = widthheight       # size to work on to get best results
    realWidthHeight = width+height      # real size of the input image
    scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling
    
    if nrAreas == 0: # auto 
        nrAreas = int(optiWidthHeight/150)
        if nrAreas < 8:
            nrAreas = 8
        if nrAreas > 32:
            nrAreas = 32
            
    suffix = ", having a knitted look, on "+str(int(nrAreas))+" colour areas"
    
    if '.' in patname:                          #1.1.1
        split = patname.split(".", 1)           #1.1.1
        pname = split[0]                        #1.1.1
    suffix = suffix+", using pattern "+pname    #1.1.1

    # Set up an undo group, so the operation will be undone in one step.
    #pdb.gimp_image_undo_group_start(inImage)
    # fix all required context settings 
    pdb.gimp_context_set_antialias(True)
    pdb.gimp_context_set_feather(False)
    pdb.gimp_context_set_feather_radius(0, 0)
    pdb.gimp_context_set_sample_merged(False) 
    pdb.gimp_context_set_sample_criterion(SELECT_CRITERION_COMPOSITE) 
    pdb.gimp_context_set_sample_threshold(0.00)
    pdb.gimp_context_set_sample_transparent(False)       # beware, if TRUE may change the area selected!
    pdb.gimp_context_set_interpolation(3)                # 1.0.1 INTERPOLATION_LANCZOS not understood by Gimp 2.9.5
    
    newwidth=int(width*scale)
    newheight=int(height*scale)
    
    # ==================================================================
    # NEW OUTPUT IMAGE
    # ------------------------------------------------------------------
    # 
    newImage = pdb.gimp_image_new (newwidth, newheight, itype)                    
    newBG = pdb.gimp_layer_new_from_drawable(inDrawable, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "newBG: Background Image Layer"
    newBG.opacity = 100.0 
    newImage.add_layer (newBG, 0)
    pdb.gimp_layer_scale(newBG, newwidth, newheight, False)
    pdb.gimp_layer_add_alpha (newBG)    #  RGBA
    # check/avoid too dark source images
    meanV = get_value (newBG)
    #info_msg ("meanV = "+str(meanV))
    if meanV > 110 and meanV < 150: # normal average 
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6, (0,32, 130,130, 255,223))
    elif meanV <= 109:  # dark
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6, (0,32, 130,150, 255,223))
    else: #light
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6, (0,32, 130,110, 255,223))
    #pdb.gimp_hue_saturation (newBG, 0, 0, 0, +60)    
    # check/avoid extremes of saturation images
    satV = get_saturation (newBG)
    #info_msg ("satV = "+str(satV))
    if satV > 160:
        pdb.gimp_hue_saturation (newBG, 0, 0, 0, -40)
    elif satV < 80:
        pdb.gimp_hue_saturation (newBG, 0, 0, 0, +40)
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    # =====================================================================
    # use work images to create the quantized layer
    # ---------------------------------------------------------------------
    #
    QUANT_IMAGE = pdb.gimp_image_duplicate (newImage)
    QUANT_DRAWABLE = QUANT_IMAGE.active_layer
    QUANT_DRAWABLE.name = "Quant BG Layer"
    
    # change image mode to indexed 
    
    pdb.gimp_image_convert_indexed (QUANT_IMAGE, 0, 0, nrAreas, 0, 0, "ignored")     
    is_indexed = pdb.gimp_drawable_is_indexed (QUANT_DRAWABLE) 
    if is_indexed == 1:
        hidden_message = "OK, converted"
    else:    
        erro_msg ("ERROR! QUANT_IMAGE not converted, exit")
        return
    # 2024 add-----------------------
    # get the number of pixels in the colormap (which exists for indexed images)
    (count, colors)=pdb.gimp_image_get_colormap(QUANT_IMAGE)
    # the count is nr_of_colors * 3, example 8*3=24
    nr_areas=(count/3) 
    #info_msg("nr of colors in the palette: "+str(nr_areas)) 
  
    temp_palette=pdb.gimp_palette_new(pdb.gimp_image_get_name(QUANT_IMAGE))
    
    index=0
    while index<count:
        # Compute the color
        color= (colors[index], colors[index+1], colors[index+2])
        # Add to the filter temporary palette
        pdb.gimp_palette_add_entry(temp_palette, str(index/3), color)
        index+=3 
    # Endwhile index<count
  
    # the temporary palette has been created            
    # ======================================

    # 2024 end
    #  
    # re-set image to RGB to proceed
    # ==============================
    pdb.gimp_image_convert_rgb (QUANT_IMAGE)
    pdb.gimp_layer_add_alpha (QUANT_DRAWABLE)                        #  RGBA
    
    # =====================================================================
    # bring the created quantized layer back to the output image
    # ---------------------------------------------------------------------
    #
    
    quantLayer = pdb.gimp_layer_new_from_drawable(QUANT_DRAWABLE, newImage)
    newImage.add_layer (quantLayer, 0)
    quantLayer.mode = LAYER_MODE_NORMAL
    quantLayer.name = "quantLayer: Quantized Image Layer"
    quantLayer.opacity = 100.0
    # delete work image
    pdb.gimp_image_delete (QUANT_IMAGE)
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    newsprintLayer = quantLayer.copy()
    newImage.add_layer (newsprintLayer, 0)
    newsprintLayer.mode = LAYER_MODE_NORMAL
    newsprintLayer.name = "newsprintLayer: News-print Image Layer"
    newsprintLayer.opacity = 100.0
    # increase contrast
    pdb.gimp_brightness_contrast (newsprintLayer, 0, +50)
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    # remove pure blacks and pure white
    pdb.gimp_curves_spline (newsprintLayer, HISTOGRAM_VALUE, 6, (0,32,128,128,255,223))
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    pdb.plug_in_newsprint (newImage, newsprintLayer,
                           5, #cell width
                           1, #color space = Separate to { GRAYSCALE (0), RGB (1), CMYK (2), LUMINANCE (3) }
                           0, #% black to pull out (CMYK only)
                           0,0, #gray angle, gray spot { DOTS (0), LINES (1), DIAMONDS (2), EUCLIDIAN-DOT (3), PS-DIAMONDS (4) }
                           5.0,1, #red angle, gray spot
                           5.0,1, #green angle, gray spot
                           5.0,1, #blue angle, gray spot
                           1)   #oversample
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    shiftedHLayer = newsprintLayer.copy()
    newImage.add_layer (shiftedHLayer, 0)
    shiftedHLayer.mode = LAYER_MODE_NORMAL
    shiftedHLayer.name = "shiftedHLayer: Shifted horizontally Image Layer"
    shiftedHLayer.opacity = 100.0 
    pdb.plug_in_shift (newImage, shiftedHLayer, 5, ORIENTATION_HORIZONTAL)
    pdb.plug_in_spread (newImage, shiftedHLayer, 1, 3)
    pdb.gimp_hue_saturation (shiftedHLayer, 0, 0, 0, +30)
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    shiftedVLayer = newsprintLayer.copy()
    newImage.add_layer (shiftedVLayer, 0)
    shiftedVLayer.mode = GRAIN_MERGE_MODE
    shiftedVLayer.name = "shiftedVLayer: Shifted vertically Image Layer"
    shiftedVLayer.opacity = 60.0 
    pdb.plug_in_shift (newImage, shiftedVLayer, 5, ORIENTATION_VERTICAL)
    pdb.plug_in_spread (newImage, shiftedVLayer, 3, 1)
    pdb.gimp_hue_saturation (shiftedVLayer, 0, 0, 0, +30)
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    #2024
    
    toFill = pdb.gimp_layer_new_from_visible (newImage, newImage, "Are to fill")
    newImage.add_layer (toFill, 0)
    
#   loop to fill the N areas with pattern                      <<<<<<<<<<< for all N areas >>>>>>>>>>>
#   =====================================

    areaNR=0
    area_color=[]   # empty list
    for areaNR in range(0,nr_areas):
        one_color=pdb.gimp_palette_entry_get_color(temp_palette, areaNR)
        area_color.append(one_color)
        pdb.gimp_context_set_sample_merged(False) 
        pdb.gimp_image_select_color(newImage, CHANNEL_OP_REPLACE, quantLayer, one_color)
        #info_msg("colour = "+str(one_color[0])+","+str(one_color[1])+","+str(one_color[2])+","+str(one_color[3])+".")
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(newImage)
        if there_is_a_selection==0:
            info_msg("area nr. "+str(areaNR)+" there_is_NO_selection")
        else:    
            #info_msg("area nr. "+str(areaNR)+" there_is_a_selection")
            #icstr(one_color[0])+","+str(one_color[1])+","+str(one_color[2])+","+str(one_color[3])+".")
            #info_msg("bounds = "+str(x1)+","+str(x2)+","+str(y1)+","+str(y2)+".")
            #info_msg ("Knitting pattern chosen="+str(patname))
            pdb.gimp_context_set_pattern (patname)  #2024
            pdb.gimp_edit_fill (toFill, PATTERN_FILL)
            # obtain HSV from Colour RGB to colorize paper
            (R,G,B,A)=one_color 
            hue,sat,val = rgb2hsv ((R,G,B,A))
            #info_msg("H,S,V = "+str(hue)+","+str(sat)+","+str(val-50))
            pdb.gimp_drawable_colorize_hsl (toFill,hue,sat,val-50)
    
    ##pdb.gimp_message("end loop")
    pdb.gimp_selection_none(newImage)
    pdb.gimp_item_set_name(toFill, "filled with a pattern ")
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    #pdb.gimp_drawable_desaturate (toFill, DESATURATE_VALUE)
    pdb.gimp_curves_spline (toFill, HISTOGRAM_VALUE, 6, (0,31, 128,128, 255,223)) # CHECK !
    #pdb.gimp_brightness_contrast (toFill, 0, +40)
    #pdb.plug_in_ripple (newImage, toFill, 20, 6, 1, 0, 1, 0, 0) # CHECVK !
        
    #pdb.gimp_display_new (newImage)
    #return    
    
    #2024
   
    topBG = quantLayer.copy()    # quantized  layer
    topBG.name = "topBG: Layer for colours"
    newImage.add_layer (topBG, 0)
    # adjust values
    #pdb.gimp_curves_spline (topBG, HISTOGRAM_VALUE, 8, (0,24,48,96,207,159,255,191))
    topBG.mode = LAYER_MODE_GRAIN_MERGE   #LINEAR_BURN  #HARDLIGHT_MODE     
    topBG.opacity = 95.0
    
    ######################################################################################################################
    
    if (doFlatten == True) :
        pdb.gimp_item_set_visible (newBG, False)
        Outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        pdb.gimp_item_set_visible (newBG, True)
    else:
        Outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (Outcome, 0)
    #2024 add some noise to final outcome
    pdb.plug_in_hsv_noise (newImage, Outcome, 3, 60, 96, 96)
    
    Outcome.name = st+" "+name+suffix   #    

    Outcome.mode = LAYER_MODE_NORMAL
    Outcome.opacity = 100.0
    
    if frameamt>0:
     # create a tiny painting frame
        frameontop = Outcome.copy()
        newImage.add_layer (frameontop, 0)
        frameontop.name = "Border on Top"
        pdb.gimp_drawable_edit_clear (frameontop)
        NR=random.randint(111111,999999)
        (R,G,B,A) = framecol
        pdb.plug_in_gmic_qt(newImage, frameontop, 1,0,
        "-v - -fx_frame_painting "+str(frameamt)
        +",0.8,6,"+str(R)+","+str(G)+","+str(B)+",0.5,400,50,10,1,0.5,"+str(NR)+",0 ")
        frameontop=newImage.layers[0]
        frameontop.mode = LAYER_MODE_NORMAL
        # align all layers to FrameLayer
        # replaced by MrQ solution	
        for layer in newImage.layers:		
            pdb.gimp_layer_set_offsets(layer,
            (pdb.gimp_image_width(newImage)-pdb.gimp_drawable_width(layer))/2,
            (pdb.gimp_image_height(newImage)-pdb.gimp_drawable_height(layer))/2)
    
    pdb.gimp_display_new(newImage)
    
    pdb.gimp_displays_flush
   
    # Close the undo group.
    #pdb.gimp_image_undo_group_end(inImage)
    
    pdb.gimp_context_pop()
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)
    
    return(Outcome)


# This is the plugin registration function
register(
    "PlainKnittedLook2024",    
    "to create a new image from the source drawable having a plain knitted look",   
    "This script creates a new image from the source drawable having a plain knitted look.",
    "Diego", 
    "Diego Nassetti ", 
    "March 2024",
    "PlainKnittedLook", 
    "RGB*", 
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "idrawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image"+
                                "\n (0=same as input)"+
                                "\n (not smaller than 2000)", 2400, (0, 6000, 200)),
      (PF_ADJUSTMENT, "colorareas", "How many colour areas"+"\n (0=auto)"+
                              "\n (8-32 recommended)", 21, (0, 80, 1)),
      (PF_PATTERN, "pattern", "Crochet pattern (choose by browsing)", "Knit_New_"),
      (PF_ADJUSTMENT, "frameamt", "Frame amount in %"+"\n (0=do not add border)",
                                    4, (0, 30, 1)),
      (PF_COLOR, "framecol", "Frame color", (225,200,120)),
      (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ], 
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),
    ],
    PlainKnittedLook2024,
    menu="<Image>/Diego/Artistic",
    )

main()
